@@ -1,10 +1,10 @@
-
 import click
 from flask import current_app
 from flask.cli import FlaskGroup, with_appcontext
 
 from alerta.app import db
 from alerta.auth.utils import generate_password_hash
+from alerta.models.enums import Scope
 from alerta.models.key import ApiKey
 from alerta.models.user import User
 
@@ -21,17 +21,19 @@ def cli():
 
 @cli.command('key', short_help='Create an admin API key')
 @click.option('--username', '-u', help='Admin user')
+@click.option('--key', '-K', help='API key (default=random UUID)')
 @click.option('--all', is_flag=True, help='Create API keys for all admins')
 @with_appcontext
-def key(username, all):
+def key(username, key, all):
     """Create an admin API key."""
     if username and username not in current_app.config['ADMIN_USERS']:
         raise click.UsageError('User {} not an admin'.format(username))
 
-    def create_key(admin):
+    def create_key(admin, key):
         key = ApiKey(
             user=admin,
-            scopes=['admin', 'write', 'read'],
+            key=key,
+            scopes=[Scope.admin, Scope.write, Scope.read],
             text='Admin key created by alertad script',
             expire_time=None
         )
@@ -45,9 +47,11 @@ def key(username, all):
 
     if all:
         for admin in current_app.config['ADMIN_USERS']:
-            create_key(admin)
+            create_key(admin, key)
+    elif username:
+        create_key(username, key)
     else:
-        create_key(username)
+        raise click.UsageError("Must set '--username' or use '--all'")
 
 
 @cli.command('keys', short_help='List admin API keys')
@@ -57,7 +61,7 @@ def keys():
     for admin in current_app.config['ADMIN_USERS']:
         try:
             db.get_db()  # init db on global app context
-            keys = ApiKey.find_by_user(admin)
+            keys = [k for k in ApiKey.find_by_user(admin) if Scope.admin in k.scopes]
         except Exception as e:
             click.echo('ERROR: {}'.format(e))
         else:
@@ -71,20 +75,24 @@ def keys():
 @click.option('--all', is_flag=True, help='Create users for all admins')
 @with_appcontext
 def user(username, password, all):
-    """Create admin users."""
+    """Create admin users (BasicAuth only)."""
+    if current_app.config['AUTH_PROVIDER'] != 'basic':
+        raise click.UsageError('Not required for {} admin users'.format(current_app.config['AUTH_PROVIDER']))
     if username and username not in current_app.config['ADMIN_USERS']:
         raise click.UsageError('User {} not an admin'.format(username))
     if not username and not all:
         raise click.UsageError('Missing option "--username".')
 
     def create_user(admin):
+        email = admin if '@' in admin else None
         user = User(
-            name=admin,
-            email=admin,
+            name='Admin user',
+            login=admin,
             password=generate_password_hash(password),
-            roles=None,
-            text='Admin user created by alertad script',
-            email_verified=True
+            roles=['admin'],
+            text='Created by alertad script',
+            email=email,
+            email_verified=bool(email)
         )
         try:
             db.get_db()  # init db on global app context
@@ -92,7 +100,7 @@ def user(username, password, all):
         except Exception as e:
             click.echo('ERROR: {}'.format(e))
         else:
-            click.echo('{} {}'.format(user.id, user.name))
+            click.echo('{} {}'.format(user.id, user.login))
 
     if all:
         for admin in current_app.config['ADMIN_USERS']:
@@ -108,13 +116,12 @@ def users():
     for admin in current_app.config['ADMIN_USERS']:
         try:
             db.get_db()  # init db on global app context
-            user = User.find_by_email(admin)
+            user = User.find_by_username(admin)
         except Exception as e:
             click.echo('ERROR: {}'.format(e))
         else:
             if user:
-                click.echo('{} {}'.format(user.id, user.name))
-
+                click.echo('{} {}'.format(user.id, user.login))
 
 
 if __name__ == '__main__':
