@@ -1,17 +1,15 @@
-
 from importlib import import_module
-
-try:
-    from urllib.parse import urlparse
-except ImportError:
-    from urlparse import urlparse
+from typing import NamedTuple
+from urllib.parse import urlparse
 
 from flask import g
 
 # http://stackoverflow.com/questions/8544983/dynamically-mixin-a-base-class-to-an-instance-in-python
 
+Query = NamedTuple('Query', [('where', str), ('sort', str), ('group', str)])
 
-class Base(object):
+
+class Base:
     pass
 
 
@@ -19,6 +17,8 @@ def get_backend(app):
     db_uri = app.config['DATABASE_URL']
     backend = urlparse(db_uri).scheme
 
+    if backend.startswith('mongodb'):
+        backend = 'mongodb'
     if backend == 'postgresql':
         backend = 'postgres'
     return backend
@@ -27,7 +27,7 @@ def get_backend(app):
 def load_backend(backend):
     try:
         return import_module('alerta.database.backends.%s' % backend)
-    except:
+    except Exception:
         raise ImportError('Failed to load %s database backend' % backend)
 
 
@@ -43,10 +43,14 @@ class Database(Base):
         cls = load_backend(backend)
         self.__class__ = type('DatabaseImpl', (cls.Backend, Database), {})
 
-        self.create_engine(app, uri=app.config['DATABASE_URL'], dbname=app.config['DATABASE_NAME'])
+        try:
+            self.create_engine(app, uri=app.config['DATABASE_URL'], dbname=app.config['DATABASE_NAME'])
+        except Exception as e:
+            if app.config['DATABASE_RAISE_ON_ERROR']:
+                raise
+            app.logger.warning(e)
 
-        app.before_request(self.get_db)
-        app.teardown_request(self.teardown_db)
+        app.teardown_appcontext(self.teardown_db)
 
     def create_engine(self, app, uri, dbname=None):
         raise NotImplementedError('Database engine has no create_engine() method')
@@ -66,21 +70,23 @@ class Database(Base):
     def is_alive(self):
         raise NotImplementedError
 
-    def close(self):
+    def close(self, db):
         raise NotImplementedError('Database engine has no close() method')
 
     def destroy(self):
         raise NotImplementedError('Database engine has no destroy() method')
 
     def get_db(self):
-        g.db = self.connect()
+        if 'db' not in g:
+            g.db = self.connect()
+        return g.db
 
     def teardown_db(self, exc):
-        db = getattr(g, 'db', None)
+        db = g.pop('db', None)
         if db is not None:
-            self.close()
+            self.close(db)
 
-    #### ALERTS
+    # ALERTS
 
     def get_severity(self, alert):
         raise NotImplementedError
@@ -88,9 +94,6 @@ class Database(Base):
     def get_status(self, alert):
         raise NotImplementedError
 
-    def get_status_and_value(self, alert):
-        raise NotImplementedError
-
     def is_duplicate(self, alert):
         raise NotImplementedError
 
@@ -109,12 +112,15 @@ class Database(Base):
     def create_alert(self, alert):
         raise NotImplementedError
 
+    def set_alert(self, id, severity, status, tags, attributes, timeout, previous_severity, update_time, history=None):
+        raise NotImplementedError
+
     def get_alert(self, id, customers=None):
         raise NotImplementedError
 
-    #### STATUS, TAGS, ATTRIBUTES
+    # STATUS, TAGS, ATTRIBUTES
 
-    def set_status(self, id, status, timeout, history=None):
+    def set_status(self, id, status, timeout, update_time, history=None):
         raise NotImplementedError
 
     def tag_alert(self, id, tags):
@@ -126,18 +132,38 @@ class Database(Base):
     def update_attributes(self, id, old_attrs, new_attrs):
         raise NotImplementedError
 
+    def add_history(self, id, history):
+        raise NotImplementedError
+
     def delete_alert(self, id):
         raise NotImplementedError
 
-    #### SEARCH & HISTORY
+    # BULK
+
+    def tag_alerts(self, query=None, tags=None):
+        raise NotImplementedError
+
+    def untag_alerts(self, query=None, tags=None):
+        raise NotImplementedError
+
+    def update_attributes_by_query(self, query=None, attributes=None):
+        raise NotImplementedError
+
+    def delete_alerts(self, query=None):
+        raise NotImplementedError
+
+    # SEARCH & HISTORY
 
     def get_alerts(self, query=None, page=None, page_size=None):
         raise NotImplementedError
 
+    def get_alert_history(self, alert, page=None, page_size=None):
+        raise NotImplementedError
+
     def get_history(self, query=None, page=None, page_size=None):
         raise NotImplementedError
 
-    #### COUNTS
+    # COUNTS
 
     def get_count(self, query=None):
         raise NotImplementedError
@@ -151,28 +177,36 @@ class Database(Base):
     def get_counts_by_status(self, query=None):
         raise NotImplementedError
 
-    def get_topn_count(self, query, group="event", topn=10):
+    def get_topn_count(self, query, group='event', topn=100):
         raise NotImplementedError
 
-    def get_topn_flapping(self, query, group="event", topn=10):
+    def get_topn_flapping(self, query, group='event', topn=100):
         raise NotImplementedError
 
-    #### ENVIRONMENTS
-
-    def get_environments(self, query=None, topn=100):
+    def get_topn_standing(self, query, group='event', topn=100):
         raise NotImplementedError
 
-    #### SERVICES
+    # ENVIRONMENTS
 
-    def get_services(self, query=None, topn=100):
+    def get_environments(self, query=None, topn=1000):
         raise NotImplementedError
 
-    #### TAGS
+    # SERVICES
 
-    def get_tags(self, query=None, topn=100):
+    def get_services(self, query=None, topn=1000):
         raise NotImplementedError
 
-    #### BLACKOUTS
+    # ALERT GROUPS
+
+    def get_alert_groups(self, query=None, topn=1000):
+        raise NotImplementedError
+
+    # ALERT TAGS
+
+    def get_alert_tags(self, query=None, topn=1000):
+        raise NotImplementedError
+
+    # BLACKOUTS
 
     def create_blackout(self, blackout):
         raise NotImplementedError
@@ -186,10 +220,13 @@ class Database(Base):
     def is_blackout_period(self, alert):
         raise NotImplementedError
 
+    def update_blackout(self, id, **kwargs):
+        raise NotImplementedError
+
     def delete_blackout(self, id):
         raise NotImplementedError
 
-    #### HEARTBEATS
+    # HEARTBEATS
 
     def upsert_heartbeat(self, heartbeat):
         raise NotImplementedError
@@ -203,24 +240,27 @@ class Database(Base):
     def delete_heartbeat(self, id):
         raise NotImplementedError
 
-    #### API KEYS
+    # API KEYS
 
     def create_key(self, key):
         raise NotImplementedError
 
-    def get_key(self, key):
+    def get_key(self, key, user=None):
         raise NotImplementedError
 
     def get_keys(self, query=None):
         raise NotImplementedError
 
+    def update_key(self, key, **kwargs):
+        raise NotImplementedError
+
     def update_key_last_used(self, key):
         raise NotImplementedError
 
     def delete_key(self, key):
         raise NotImplementedError
 
-    #### USERS
+    # USERS
 
     def create_user(self, user):
         raise NotImplementedError
@@ -231,6 +271,9 @@ class Database(Base):
     def get_users(self, query=None):
         raise NotImplementedError
 
+    def get_user_by_username(self, username):
+        raise NotImplementedError
+
     def get_user_by_email(self, email):
         raise NotImplementedError
 
@@ -240,9 +283,6 @@ class Database(Base):
     def update_last_login(self, id):
         raise NotImplementedError
 
-    def set_email_hash(self, id, hash):
-        raise NotImplementedError
-
     def update_user(self, id, **kwargs):
         raise NotImplementedError
 
@@ -252,7 +292,39 @@ class Database(Base):
     def delete_user(self, id):
         raise NotImplementedError
 
-    #### PERMISSIONS
+    def set_email_hash(self, id, hash):
+        raise NotImplementedError
+
+    # GROUPS
+
+    def create_group(self, group):
+        raise NotImplementedError
+
+    def get_group(self, id):
+        raise NotImplementedError
+
+    def get_group_users(self, id):
+        raise NotImplementedError
+
+    def get_groups(self, query=None):
+        raise NotImplementedError
+
+    def update_group(self, id, **kwargs):
+        raise NotImplementedError
+
+    def add_user_to_group(self, group, user):
+        raise NotImplementedError
+
+    def remove_user_from_group(self, group, user):
+        raise NotImplementedError
+
+    def delete_group(self, id):
+        raise NotImplementedError
+
+    def get_groups_by_user(self, user):
+        raise NotImplementedError
+
+    # PERMISSIONS
 
     def create_perm(self, perm):
         raise NotImplementedError
@@ -263,13 +335,16 @@ class Database(Base):
     def get_perms(self, query=None):
         raise NotImplementedError
 
+    def update_perm(self, id, **kwargs):
+        raise NotImplementedError
+
     def delete_perm(self, id):
         raise NotImplementedError
 
     def get_scopes_by_match(self, login, matches):
         raise NotImplementedError
 
-    #### CUSTOMERS
+    # CUSTOMERS
 
     def create_customer(self, customer):
         raise NotImplementedError
@@ -280,13 +355,16 @@ class Database(Base):
     def get_customers(self, query=None):
         raise NotImplementedError
 
+    def update_customer(self, id, **kwargs):
+        raise NotImplementedError
+
     def delete_customer(self, id):
         raise NotImplementedError
 
     def get_customers_by_match(self, login, matches):
         raise NotImplementedError
 
-    #### METRICS
+    # METRICS
 
     def get_metrics(self, type=None):
         raise NotImplementedError
@@ -300,7 +378,7 @@ class Database(Base):
     def update_timer(self, timer):
         raise NotImplementedError
 
-    #### HOUSEKEEPING
+    # HOUSEKEEPING
 
     def housekeeping(self, expired_threshold, info_threshold):
         raise NotImplementedError
@@ -319,7 +397,7 @@ class QueryBuilder(Base):
         self.__class__ = type('QueryBuilderImpl', (cls.QueryBuilderImpl, QueryBuilder), {})
 
     @staticmethod
-    def from_params(params, query_time=None):
+    def from_params(params, customers=None, query_time=None):
         raise NotImplementedError('QueryBuilder has no from_params() method')
 
     @staticmethod
