@@ -1,22 +1,27 @@
-
 import os
 import platform
 import sys
 from datetime import datetime
+from typing import Optional  # noqa
+from typing import Any, Dict, List, Tuple, Union
 from uuid import uuid4
 
-from flask import current_app
+from flask import current_app, g
 
-from alerta.app import db, severity
-from alerta.models import status_code
+from alerta.app import alarm_model, db
+from alerta.database.base import Query
 from alerta.models.history import History, RichHistory
-from alerta.utils.api import absolute_url
 from alerta.utils.format import DateTime
+from alerta.utils.hooks import status_change_hook
+from alerta.utils.response import absolute_url
+
+JSON = Dict[str, Any]
+NoneType = type(None)
 
 
-class Alert(object):
+class Alert:
 
-    def __init__(self, resource, event, **kwargs):
+    def __init__(self, resource: str, event: str, **kwargs) -> None:
 
         if not resource:
             raise ValueError('Missing mandatory value for "resource"')
@@ -27,26 +32,37 @@ class Alert(object):
             raise ValueError('Attribute keys must not contain "." or "$"')
         if isinstance(kwargs.get('value', None), int):
             kwargs['value'] = str(kwargs['value'])
+        for attr in ['create_time', 'receive_time', 'last_receive_time']:
+            if not isinstance(kwargs.get(attr), (datetime, NoneType)):  # type: ignore
+                raise ValueError("Attribute '{}' must be datetime type".format(attr))
 
-        self.id = kwargs.get('id', str(uuid4()))
+        timeout = kwargs.get('timeout') if kwargs.get('timeout') is not None else current_app.config['ALERT_TIMEOUT']
+        try:
+            timeout = int(timeout)  # type: ignore
+        except ValueError:
+            raise ValueError("Could not convert 'timeout' value of '{}' to an integer".format(timeout))
+        if timeout < 0:
+            raise ValueError("Invalid negative 'timeout' value ({})".format(timeout))
+
+        self.id = kwargs.get('id', None) or str(uuid4())
         self.resource = resource
         self.event = event
-        self.environment = kwargs.get('environment', None) or ""
-        self.severity = kwargs.get('severity', None) or current_app.config['DEFAULT_NORMAL_SEVERITY']
+        self.environment = kwargs.get('environment', None) or ''
+        self.severity = kwargs.get('severity', None) or alarm_model.DEFAULT_NORMAL_SEVERITY
         self.correlate = kwargs.get('correlate', None) or list()
         if self.correlate and event not in self.correlate:
             self.correlate.append(event)
-        self.status = kwargs.get('status', None) or status_code.UNKNOWN
+        self.status = kwargs.get('status', None) or alarm_model.DEFAULT_STATUS
         self.service = kwargs.get('service', None) or list()
-        self.group = kwargs.get('group', None) or "Misc"
+        self.group = kwargs.get('group', None) or 'Misc'
         self.value = kwargs.get('value', None)
-        self.text = kwargs.get('text', None) or ""
+        self.text = kwargs.get('text', None) or ''
         self.tags = kwargs.get('tags', None) or list()
         self.attributes = kwargs.get('attributes', None) or dict()
-        self.origin = kwargs.get('origin', None) or '%s/%s' % (os.path.basename(sys.argv[0]), platform.uname()[1])
-        self.event_type = kwargs.get('event_type', kwargs.get('type', None)) or "exceptionAlert"
+        self.origin = kwargs.get('origin', None) or '{}/{}'.format(os.path.basename(sys.argv[0]), platform.uname()[1])
+        self.event_type = kwargs.get('event_type', kwargs.get('type', None)) or 'exceptionAlert'
         self.create_time = kwargs.get('create_time', None) or datetime.utcnow()
-        self.timeout = kwargs.get('timeout', None) or current_app.config['ALERT_TIMEOUT']
+        self.timeout = timeout
         self.raw_data = kwargs.get('raw_data', None)
         self.customer = kwargs.get('customer', None)
 
@@ -57,10 +73,11 @@ class Alert(object):
         self.receive_time = kwargs.get('receive_time', None) or datetime.utcnow()
         self.last_receive_id = kwargs.get('last_receive_id', None)
         self.last_receive_time = kwargs.get('last_receive_time', None)
+        self.update_time = kwargs.get('update_time', None)
         self.history = kwargs.get('history', None) or list()
 
     @classmethod
-    def parse(cls, json):
+    def parse(cls, json: JSON) -> 'Alert':
         if not isinstance(json.get('correlate', []), list):
             raise ValueError('correlate must be a list')
         if not isinstance(json.get('service', []), list):
@@ -71,8 +88,11 @@ class Alert(object):
             raise ValueError('attributes must be a JSON object')
         if not isinstance(json.get('timeout') if json.get('timeout', None) is not None else 0, int):
             raise ValueError('timeout must be an integer')
+        if json.get('customer', None) == '':
+            raise ValueError('customer must not be an empty string')
 
         return Alert(
+            id=json.get('id', None),
             resource=json.get('resource', None),
             event=json.get('event', None),
             environment=json.get('environment', None),
@@ -87,14 +107,14 @@ class Alert(object):
             attributes=json.get('attributes', dict()),
             origin=json.get('origin', None),
             event_type=json.get('type', None),
-            create_time=DateTime.parse(json.get('createTime')),
+            create_time=DateTime.parse(json['createTime']) if 'createTime' in json else None,
             timeout=json.get('timeout', None),
             raw_data=json.get('rawData', None),
             customer=json.get('customer', None)
         )
 
     @property
-    def serialize(self):
+    def serialize(self) -> Dict[str, Any]:
         return {
             'id': self.id,
             'href': absolute_url('/alert/' + self.id),
@@ -123,27 +143,28 @@ class Alert(object):
             'receiveTime': self.receive_time,
             'lastReceiveId': self.last_receive_id,
             'lastReceiveTime': self.last_receive_time,
-            'history': [h.serialize for h in self.history]
+            'updateTime': self.update_time,
+            'history': [h.serialize for h in sorted(self.history, key=lambda x: x.update_time)]
         }
 
-    def get_id(self, short=False):
+    def get_id(self, short: bool = False) -> str:
         return self.id[:8] if short else self.id
 
-    def get_body(self, history=True):
+    def get_body(self, history: bool = True) -> Dict[str, Any]:
         body = self.serialize
         body.update({
-            key: DateTime.iso8601(body[key]) for key in ['createTime', 'lastReceiveTime', 'receiveTime']
+            key: DateTime.iso8601(body[key]) for key in ['createTime', 'lastReceiveTime', 'receiveTime', 'updateTime'] if body[key]
         })
         if not history:
             body['history'] = []
         return body
 
-    def __repr__(self):
-        return 'Alert(id=%r, environment=%r, resource=%r, event=%r, severity=%r, status=%r, customer=%r)' % (
+    def __repr__(self) -> str:
+        return 'Alert(id={!r}, environment={!r}, resource={!r}, event={!r}, severity={!r}, status={!r}, customer={!r})'.format(
             self.id, self.environment, self.resource, self.event, self.severity, self.status, self.customer)
 
     @classmethod
-    def from_document(cls, doc):
+    def from_document(cls, doc: Dict[str, Any]) -> 'Alert':
         return Alert(
             id=doc.get('id', None) or doc.get('_id'),
             resource=doc.get('resource', None),
@@ -171,11 +192,12 @@ class Alert(object):
             receive_time=doc.get('receiveTime', None),
             last_receive_id=doc.get('lastReceiveId', None),
             last_receive_time=doc.get('lastReceiveTime', None),
+            update_time=doc.get('updateTime', None),
             history=[History.from_db(h) for h in doc.get('history', list())]
         )
 
     @classmethod
-    def from_record(cls, rec):
+    def from_record(cls, rec) -> 'Alert':
         return Alert(
             id=rec.id,
             resource=rec.resource,
@@ -203,76 +225,111 @@ class Alert(object):
             receive_time=rec.receive_time,
             last_receive_id=rec.last_receive_id,
             last_receive_time=rec.last_receive_time,
+            update_time=getattr(rec, 'update_time'),
             history=[History.from_db(h) for h in rec.history]
         )
 
     @classmethod
-    def from_db(cls, r):
+    def from_db(cls, r: Union[Dict, Tuple]) -> 'Alert':
         if isinstance(r, dict):
             return cls.from_document(r)
         elif isinstance(r, tuple):
             return cls.from_record(r)
 
-    def is_duplicate(self):
-        return db.is_duplicate(self)
+    def is_duplicate(self) -> Optional['Alert']:
+        """Return duplicate alert or None"""
+        return Alert.from_db(db.is_duplicate(self))
 
-    def is_correlated(self):
-        return db.is_correlated(self)
+    def is_correlated(self) -> Optional['Alert']:
+        """Return correlated alert or None"""
+        return Alert.from_db(db.is_correlated(self))
 
-    def is_flapping(self, window=1800, count=2):
+    def is_flapping(self, window: int = 1800, count: int = 2) -> bool:
         return db.is_flapping(self, window, count)
 
+    def get_status_and_value(self):
+        return [(h.status, h.value) for h in self.get_alert_history(self, page=1, page_size=10) if h.status]
+
+    def _get_hist_info(self, action=None):
+        h_loop = self.get_alert_history(alert=self)
+        if len(h_loop) == 1:
+            return h_loop[0].status, h_loop[0].value, None
+        if action == 'unack':
+            find = 'ack'
+        elif action == 'unshelve':
+            find = 'shelve'
+        else:
+            find = None
+        for h, h_next in zip(h_loop, h_loop[1:]):
+            if not find or h.change_type == find:
+                return h_loop[0].status, h_loop[0].value, h_next.status
+        return None, None, None
+
     # de-duplicate an alert
-    def deduplicate(self):
+    def deduplicate(self, duplicate_of) -> 'Alert':
         now = datetime.utcnow()
 
-        previous_status, previous_value = db.get_status_and_value(self)
-        self.status = status_code.status_from_severity(
-            previous_severity=self.severity,
-            current_severity=self.severity,
-            previous_status=previous_status,
-            current_status=previous_status
+        status, previous_value, previous_status = self._get_hist_info()
+
+        _, new_status = alarm_model.transition(
+            alert=self,
+            current_status=status,
+            previous_status=previous_status
         )
 
         self.repeat = True
         self.last_receive_id = self.id
         self.last_receive_time = now
 
-        if self.status != previous_status:
-            history = History(
-                id=self.id,
-                event=self.event,
-                status=self.status,
-                text="duplicate alert with status change",
-                change_type="status",
-                update_time=now
-            )
-        elif self.value != previous_value:
+        if new_status != status:
+            text = 'duplicate alert (with status change)'
+            r = status_change_hook.send(duplicate_of, status=new_status, text=text)
+            _, (_, new_status, text) = r[0]
+            self.update_time = now
+
             history = History(
                 id=self.id,
                 event=self.event,
+                severity=self.severity,
+                status=new_status,
                 value=self.value,
-                text="duplicate alert with value change",
-                change_type="value",
-                update_time=now
+                text=text,
+                change_type='status',
+                update_time=self.create_time,
+                user=g.login,
+            )  # type: Optional[History]
+
+        elif current_app.config['HISTORY_ON_VALUE_CHANGE'] and self.value != previous_value:
+            history = History(
+                id=self.id,
+                event=self.event,
+                severity=self.severity,
+                status=status,
+                value=self.value,
+                text='duplicate alert (with value change)',
+                change_type='value',
+                update_time=self.create_time,
+                user=g.login
             )
         else:
             history = None
+
+        self.status = new_status
         return Alert.from_db(db.dedup_alert(self, history))
 
     # correlate an alert
-    def update(self):
+    def update(self, correlate_with) -> 'Alert':
         now = datetime.utcnow()
 
         self.previous_severity = db.get_severity(self)
-        previous_status = db.get_status(self)
-        self.trend_indication = severity.trend(self.previous_severity, self.severity)
+        self.trend_indication = alarm_model.trend(self.previous_severity, self.severity)
 
-        self.status = status_code.status_from_severity(
-            previous_severity=self.previous_severity,
-            current_severity=self.severity,
-            previous_status=previous_status,
-            current_status=self.status
+        status, _, previous_status = self._get_hist_info()
+
+        _, new_status = alarm_model.transition(
+            alert=self,
+            current_status=status,
+            previous_status=previous_status
         )
 
         self.duplicate_count = 0
@@ -280,172 +337,305 @@ class Alert(object):
         self.receive_time = now
         self.last_receive_id = self.id
         self.last_receive_time = now
+        text = 'correlated alert'
+
+        if new_status != status:
+            r = status_change_hook.send(correlate_with, status=new_status, text=text)
+            _, (_, new_status, text) = r[0]
+            self.update_time = now
 
         history = [History(
             id=self.id,
             event=self.event,
             severity=self.severity,
+            status=new_status,
             value=self.value,
-            text=self.text,
-            change_type="severity",
-            update_time=now
+            text=text,
+            change_type='severity',
+            update_time=self.create_time,
+            user=g.login
         )]
 
-        if self.status != previous_status:
-            history.append(History(
-                id=self.id,
-                event=self.event,
-                status=self.status,
-                text="correlated alert status change",
-                change_type="status",
-                update_time=now
-            ))
-
+        self.status = new_status
         return Alert.from_db(db.correlate_alert(self, history))
 
     # create an alert
-    def create(self):
-        if self.status == status_code.UNKNOWN:
-            self.status = status_code.status_from_severity(
-                previous_severity=current_app.config['DEFAULT_PREVIOUS_SEVERITY'],
-                current_severity=self.severity
-            )
-        trend_indication = severity.trend(current_app.config['DEFAULT_PREVIOUS_SEVERITY'], self.severity)
+    def create(self) -> 'Alert':
+        now = datetime.utcnow()
+
+        trend_indication = alarm_model.trend(alarm_model.DEFAULT_PREVIOUS_SEVERITY, self.severity)
+
+        _, self.status = alarm_model.transition(
+            alert=self
+        )
 
         self.duplicate_count = 0
         self.repeat = False
-        self.previous_severity = current_app.config['DEFAULT_PREVIOUS_SEVERITY']
+        self.previous_severity = alarm_model.DEFAULT_PREVIOUS_SEVERITY
         self.trend_indication = trend_indication
-        self.receive_time = datetime.utcnow()
+        self.receive_time = now
         self.last_receive_id = self.id
-        self.last_receive_time = self.receive_time
+        self.last_receive_time = now
+        self.update_time = now
 
         self.history = [History(
             id=self.id,
             event=self.event,
             severity=self.severity,
-            value=self.value,
-            text=self.text,
-            change_type='severity',
-            update_time=self.create_time
-        )]
-
-        self.history.append(History(
-            id=self.id,
-            event=self.event,
             status=self.status,
-            text="new alert status change",
-            change_type='status',
-            update_time=self.last_receive_time
-        ))
+            value=self.value,
+            text='new alert',
+            change_type='new',
+            update_time=self.create_time,
+            user=g.login
+        )]
 
         return Alert.from_db(db.create_alert(self))
 
     # retrieve an alert
     @staticmethod
-    def find_by_id(id, customers=None):
+    def find_by_id(id: str, customers: List[str] = None) -> 'Alert':
         return Alert.from_db(db.get_alert(id, customers))
 
-    def is_blackout(self):
+    def is_blackout(self) -> bool:
+        """Does this alert match a blackout period?"""
         if not current_app.config['NOTIFICATION_BLACKOUT']:
             if self.severity in current_app.config['BLACKOUT_ACCEPT']:
                 return False
         return db.is_blackout_period(self)
 
+    @property
+    def is_suppressed(self) -> bool:
+        """Is the alert status 'blackout'?"""
+        return alarm_model.is_suppressed(self)
+
     # set alert status
-    def set_status(self, status, text='', timeout=None):
-        self.timeout = timeout or current_app.config['ALERT_TIMEOUT']
+    def set_status(self, status: str, text: str = '', timeout: int = None) -> 'Alert':
+        now = datetime.utcnow()
+
+        timeout = timeout or current_app.config['ALERT_TIMEOUT']
         history = History(
             id=self.id,
             event=self.event,
+            severity=self.severity,
             status=status,
+            value=self.value,
             text=text,
-            change_type="status",
-            update_time=datetime.utcnow()
+            change_type='status',
+            update_time=now,
+            user=g.login
         )
-        return db.set_status(self.id, status, timeout, history)
+        return db.set_status(self.id, status, timeout, update_time=now, history=history)
 
     # tag an alert
-    def tag(self, tags):
+    def tag(self, tags: List[str]) -> bool:
         return db.tag_alert(self.id, tags)
 
     # untag an alert
-    def untag(self, tags):
+    def untag(self, tags: List[str]) -> bool:
         return db.untag_alert(self.id, tags)
 
     # update alert attributes
-    def update_attributes(self, attributes):
+    def update_attributes(self, attributes: Dict[str, Any]) -> bool:
         return db.update_attributes(self.id, self.attributes, attributes)
 
+    # add note
+    def add_note(self, note: str) -> bool:
+        history = History(
+            id=self.id,
+            event=self.event,
+            severity=self.severity,
+            status=self.status,
+            value=self.value,
+            text=note,
+            change_type='note',
+            update_time=datetime.utcnow(),
+            user=g.login
+        )
+        return db.add_history(self.id, history)
+
     # delete an alert
-    def delete(self):
+    def delete(self) -> bool:
         return db.delete_alert(self.id)
 
+    # bulk tag
+    @staticmethod
+    def tag_find_all(query, tags):
+        return db.tag_alerts(query, tags)
+
+    # bulk untag
+    @staticmethod
+    def untag_find_all(query, tags):
+        return db.untag_alerts(query, tags)
+
+    # bulk update attributes
+    @staticmethod
+    def update_attributes_find_all(query, attributes):
+        return db.update_attributes_by_query(query, attributes)
+
+    # bulk delete
+    @staticmethod
+    def delete_find_all(query=None):
+        return db.delete_alerts(query)
+
     # search alerts
     @staticmethod
-    def find_all(query=None, page=1, page_size=100):
+    def find_all(query: Query = None, page: int = 1, page_size: int = 1000) -> List['Alert']:
         return [Alert.from_db(alert) for alert in db.get_alerts(query, page, page_size)]
 
+    @staticmethod
+    def get_alert_history(alert, page=1, page_size=100):
+        return [RichHistory.from_db(hist) for hist in db.get_alert_history(alert, page, page_size)]
+
     # list alert history
     @staticmethod
-    def get_history(query=None, page=1, page_size=100):
+    def get_history(query: Query = None, page=1, page_size=1000) -> List[RichHistory]:
         return [RichHistory.from_db(hist) for hist in db.get_history(query, page, page_size)]
 
     # get total count
     @staticmethod
-    def get_count(query=None):
+    def get_count(query: Query = None) -> Dict[str, Any]:
         return db.get_count(query)
 
     # get severity counts
     @staticmethod
-    def get_counts_by_severity(query=None):
+    def get_counts_by_severity(query: Query = None) -> Dict[str, Any]:
         return db.get_counts_by_severity(query)
 
     # get status counts
     @staticmethod
-    def get_counts_by_status(query=None):
+    def get_counts_by_status(query: Query = None) -> Dict[str, Any]:
         return db.get_counts_by_status(query)
 
     # top 10 alerts
     @staticmethod
-    def get_top10_count(query=None):
-        return db.get_topn_count(query)
+    def get_top10_count(query: Query = None) -> List[Dict[str, Any]]:
+        return db.get_topn_count(query, topn=10)
 
     # top 10 flapping
     @staticmethod
-    def get_top10_flapping(query=None):
-        return db.get_topn_flapping(query)
+    def get_top10_flapping(query: Query = None) -> List[Dict[str, Any]]:
+        return db.get_topn_flapping(query, topn=10)
+
+    # top 10 standing
+    @staticmethod
+    def get_top10_standing(query: Query = None) -> List[Dict[str, Any]]:
+        return db.get_topn_standing(query, topn=10)
 
     # get environments
     @staticmethod
-    def get_environments(query=None):
+    def get_environments(query: Query = None) -> List[str]:
         return db.get_environments(query)
 
     # get services
     @staticmethod
-    def get_services(query=None):
+    def get_services(query: Query = None) -> List[str]:
         return db.get_services(query)
 
+    # get groups
+    @staticmethod
+    def get_groups(query: Query = None) -> List[str]:
+        return db.get_alert_groups(query)
+
     # get tags
     @staticmethod
-    def get_tags(query=None):
-        return db.get_tags(query)
+    def get_tags(query: Query = None) -> List[str]:
+        return db.get_alert_tags(query)
 
     @staticmethod
-    def housekeeping(expired_threshold=2, info_threshold=12):
-        for (id, event, status, last_receive_id) in db.housekeeping(expired_threshold, info_threshold):
-            if status == 'open':
-                text = 'unshelved after timeout'
-            elif status == 'expired':
-                text = 'expired after timeout'
-            else:
-                text = 'alert timeout status change'
+    def housekeeping(expired_threshold: int = 2, info_threshold: int = 12) -> None:
+        now = datetime.utcnow()
+        expired, unshelved = db.housekeeping(expired_threshold, info_threshold)
+
+        for (id, event, last_receive_id) in expired:
             history = History(
                 id=last_receive_id,
                 event=event,
-                status=status,
-                text=text,
-                change_type="status",
-                update_time=datetime.utcnow()
+                status='expired',
+                text='auto-expired after timeout',
+                change_type='status',
+                update_time=now,
+                user=g.login
             )
-            db.set_status(id, status, timeout=current_app.config['ALERT_TIMEOUT'], history=history)
+            db.set_status(id, 'expired', timeout=current_app.config['ALERT_TIMEOUT'], update_time=now, history=history)
+
+        for (id, event, last_receive_id) in unshelved:
+            # as per ISA 18.2 recommendation 11.7.3 auto-unshelved alarms transition to open, not previous status
+            history = History(
+                id=last_receive_id,
+                event=event,
+                status='open',
+                text='auto-unshelved after timeout',
+                change_type='status',
+                update_time=now,
+                user=g.login
+            )
+            db.set_status(id, 'open', timeout=current_app.config['ALERT_TIMEOUT'], update_time=now, history=history)
+
+    def from_status(self, status: str, text: str = '', timeout: int = None) -> 'Alert':
+        now = datetime.utcnow()
+
+        self.timeout = timeout or current_app.config['ALERT_TIMEOUT']
+        history = [History(
+            id=self.id,
+            event=self.event,
+            severity=self.severity,
+            status=status,
+            value=self.value,
+            text=text,
+            change_type='status',
+            update_time=now,
+            user=g.login
+        )]
+        return Alert.from_db(db.set_alert(
+            id=self.id,
+            severity=self.severity,
+            status=status,
+            tags=self.tags,
+            attributes=self.attributes,
+            timeout=timeout,
+            previous_severity=self.previous_severity,
+            update_time=now,
+            history=history)
+        )
+
+    def from_action(self, action: str, text: str = '', timeout: int = None) -> 'Alert':
+        now = datetime.utcnow()
+
+        self.timeout = timeout or current_app.config['ALERT_TIMEOUT']
+
+        status, _, previous_status = self._get_hist_info(action)
+
+        new_severity, new_status = alarm_model.transition(
+            alert=self,
+            current_status=status,
+            previous_status=previous_status,
+            action=action
+        )
+
+        r = status_change_hook.send(self, status=new_status, text=text)
+        _, (_, new_status, text) = r[0]
+
+        history = [History(
+            id=self.id,
+            event=self.event,
+            severity=new_severity,
+            status=new_status,
+            value=self.value,
+            text=text,
+            change_type=action,
+            update_time=now,
+            user=g.login
+        )]
+
+        return Alert.from_db(db.set_alert(
+            id=self.id,
+            severity=new_severity,
+            status=new_status,
+            tags=self.tags,
+            attributes=self.attributes,
+            timeout=timeout,
+            previous_severity=self.severity if new_severity != self.severity else self.previous_severity,
+            update_time=now,
+            history=history)
+        )
